import socket
import struct
import threading

class DataType:
    ACK=1
    DATA=2
    DATA_LOW_LATENCY=3
    DATA_WITH_ACK=4


class NetworkAL(object):
    """
    Alternate implementation of the ARNetworkAL protocol, for Wifi devices.

    This implementations is fully compliant with the protocol, and has no major
    limiations.

    This implementation uses a thread to do background reads from the socket, and
    send data to the application through a listener. This listener must implement a
    'data_received' function, which will receive the following arguments:
    - type : The type of data received (ack, data, low latency, data with ack)
    - buf : The buffer on which this data was retrieved
    - seq : The sequence number of the data
    - recv_data : The actual data, as a packed string (use the struct module to unpack)
    And a 'did_disconnect' function, without arguments, which will be called if the product
    does not send any data on the network (probably because we lost the network link, or
    because the product has run out of battery)
    """

    def __init__(self, ip, c2d_port, d2c_port, listener):
        """
        Create and start a new instance of ARNetworkAL.

        Arguments:
        - ip (string) : The device address
        - c2d_port : The remove reading port
        - d2c_port : The local reading port
        - listener : A listener which will have its data_received function called
                     when a data is received from the network.
        """
        self._ip = ip
        self._c2d_port = int(c2d_port)
        self._d2c_port = int(d2c_port)
        self._listener = listener
        self._alive = False
        self._running = False
        self._thread = None
        self.start()

    def stop(self):
        """
        Stop the current ARNetworkAL instance.

        Once stopped, an instance can be restarded with the start method.
        """
        if self._running:
            self._alive = False
            self._send_sock.close()

    def start(self):
        """
        Start the current ARNetworkAL instance.

        This function has no effect if the instance is already started.
        """
        if self._running:
            return
        self._alive = True
        self._send_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._recv_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._recv_sock.settimeout(5.0)
        self._recv_sock.bind(('0.0.0.0', self._d2c_port))
        self._thread = threading.Thread(target=self._read_loop)
        self._thread.start()
        self._running = True


    def send_data(self, type, buf, seq, data):
        """
        Send the given data to the remote ARNetworkAL.

        This function returns a boolean indicating whether the send worked.
        This boolean is not an acknowlege, just an indicator that the socket
        write did not fail.

        Arguments:
        - type : The type of data (ack, data, low latency, data with ack)
        - buf : The target buffer for the data
        - seq : The sequence number of the data
        - data : The actual data (ususally a string packed with the struct module)
        """
        sock_data = struct.pack('<BBBI', type, buf, seq, len(data) + 7)
        sock_data += data
        try:
            self._send_sock.sendto(sock_data, (self._ip, self._c2d_port))
        except:
            return False
        return True

    def _read_loop(self):
        while self._alive:
            try:
                sock_data, _ = self._recv_sock.recvfrom(66000)
            except Exception as e:
                break

            the_data = sock_data
            while the_data:
                (type, buf, seq, size) = struct.unpack('<BBBI', the_data[0:7])
                recv_data = the_data[7:size]
                self._listener.data_received(type, buf, seq, recv_data)
                the_data = the_data[size:]

        self._recv_sock.close()
        self._listener.did_disconnect()
        self._running = False



import Bybop_NetworkAL
import struct
import threading

class NetworkStatus:
    OK = 0
    ERROR = 1
    TIMEOUT = 2

class Network(object):
    """
    Simple implementation of the ARNetwork protocol.

    This implementation does not support intenal fifos. If multiple threads tries to send data on the
    same buffer at the same time, the actual send order is undefined.

    The 'send_data' call is blocking to allow simpler implementation, but is not doing busy waiting so
    it can be called from a thread without locking the GIL in python implementations that use one.

    This implementation use a listener to warn the application of newly received data. The listener
    should implement a 'data_received' function accepting the following arguments:
    - buf : The buffer on which this data was retrieved
    - recv_data : The actual data, as a packed string (use the struct module to unpack)
    And a 'did_disconnect' function, without arguments, which will be called if the product
    does not send any data on the network (probably because we lost the network link, or
    because the product has run out of battery)
    """

    def __init__(self, ip, c2d_port, d2c_port, send_buffers, recv_buffers, listener):
        """
        Create a new instance of ARNetwork.

        The instance will manage internally its ARNetworkAL backend.

        Arguments:
        - ip (string) : The device address
        - c2d_port : The remove reading port
        - d2c_port : The local reading port
        - send_buffers : List of buffers which should accept data from the application
                       (i.e. which will be given to the send_data function)
        - recv_buffers : List of buffers which should accept incoming data
        """
        self._netal = Bybop_NetworkAL.NetworkAL(ip, c2d_port, d2c_port, self)
        self._listener = listener
        self._send_buffers = list(send_buffers) # The application writed to these (send to network)
        self._recv_buffers = list(recv_buffers) # The application reads from these (read from network)
        self._send_seq = {}
        self._recv_seq = {}
        self._ack_events = {}
        self._ack_seq = {}
        self._buf_locks = {}
        self._ack_events_lock = threading.Lock()

        for sndb in self._send_buffers:
            self._send_seq[sndb] = 0
            self._buf_locks[sndb] = threading.Lock()
            self._ack_events[sndb] = threading.Event()
            self._ack_seq[sndb] = 0
        for rcvb in self._recv_buffers:
            self._recv_seq[rcvb] = 255



    def stop(self):
        """
        Stop the ARNetwork instance.

        This also stops the ARNetworkAL backend.

        This function has no effect on a stopped instance.
        """
        self._netal.stop()

    def restart(self):
        """
        Restart the ARNetwork instance.

        This also restarts the ARNetworkAL backend.

        This function has no effect on a started instance.
        """
        self._netal.start()

    def _get_seq(self, buf):
        if not buf in self._send_seq:
            self._send_seq[buf] = 0
        ret = self._send_seq[buf]
        self._send_seq[buf] += 1
        self._send_seq[buf] %= 256
        return ret

    def send_data(self, buf, data, type, timeout=0.15, tries=5):
        """
        Send some data over the network, and return an ARNetworkStatus.

        The keyword arguments are only used for acknowledged data.
        For other data, the timeout is irrelevant, and only one try will be made.

        For acknowledged data, this function will block until either the acknowledge is received,
        or all the tries have been consumed in timeouts. For other data, this function returns
        almost immediately.

        Arguments:
        - buf : The target buffer for the data (must be part of the send_buffers list given to __init__)
        - data : The data to send
        - type : The type of the data (needs ack or not)

        Keyword arguments:
        - timeout : Timeout in floating point number of seconds, or None if no timeout (default 0.15)
        - tries : Total number of tries before considering a data as lost (default 5)
        """
        if not buf in self._send_buffers:
            return NetworkStatus.ERROR

        seqnum = self._get_seq(buf)
        needack = type == Bybop_NetworkAL.DataType.DATA_WITH_ACK
        status = NetworkStatus.TIMEOUT

        with self._buf_locks[buf]:

            # If we need an ack, clear any pending ack event, and set the requested seqnum
            if needack:
                with self._ack_events_lock:
                    self._ack_events[buf].clear()
                    self._ack_seq[buf] = seqnum

            # Try 'retries' times in case of timeouts
            while tries > 0 and status == NetworkStatus.TIMEOUT:
                tries -= 1

                status = NetworkStatus.OK if self._netal.send_data(type, buf, seqnum, data) else NetworkStatus.ERROR
                # We only set TIMEOUT status for acknowledged data
                if needack and status == NetworkStatus.OK: # Data with ack properly sent
                    status = NetworkStatus.OK if self._ack_events[buf].wait(timeout) else NetworkStatus.TIMEOUT
        return status

    def _send_ack(self, buf, seq):
        answer = struct.pack('<B', seq)
        abuf = buf + 128
        self._netal.send_data(Bybop_NetworkAL.DataType.ACK, abuf, self._get_seq(abuf), answer)

    def _send_pong(self, data):
        self._netal.send_data(Bybop_NetworkAL.DataType.DATA, 1, self._get_seq(1), data)

    def _should_accept(self, buf, seq):
        if not buf in self._recv_seq:
            return False

        prev = self._recv_seq[buf]
        diff = seq - prev
        ok = diff >= 0 or diff <= -10

        if ok:
            self._recv_seq[buf] = seq
        return ok


    def data_received(self, type, buf, seq, recv_data):
        """
        Implementation of the NetworkAL listener.

        This function should not be called direcly by application code !
        """
        if buf == 0: # This is a ping, send a pong !
            self._send_pong(recv_data)

        if type == Bybop_NetworkAL.DataType.ACK:
            ackbuf = buf - 128
            if ackbuf in self._send_buffers:
                seq = struct.unpack('<B', recv_data)[0]
                with self._ack_events_lock:
                    if seq == self._ack_seq[ackbuf]:
                        self._ack_events[ackbuf].set()
        elif type == Bybop_NetworkAL.DataType.DATA:
            self._process_data(buf, seq, recv_data)
        elif type == Bybop_NetworkAL.DataType.DATA_LOW_LATENCY:
            self._process_data(buf, seq, recv_data)
        elif type == Bybop_NetworkAL.DataType.DATA_WITH_ACK:
            self._process_data(buf, seq, recv_data)
            # And send ack !
            self._send_ack(buf, seq)

    def _process_data(self, buf, seq, recv_data):
        if self._should_accept(buf, seq):
            self._listener.data_received(buf, recv_data)

    def did_disconnect(self):
        """
        Implementation of the NetworkAL listener.

        This function should not be called directly by application code !
        """
        self._listener.did_disconnect()



#!/usr/bin/env python
# -*- coding:utf-8 -*-

from setuptools import setup, find_packages

setup(
        name="bybop",
        version="0.0.3",
        packages=find_packages('src'),
        package_dir={'': 'src'},
        package_data={
                        'arsdk-xml': ['xml/*.xml']
                     },
        zip_safe=False,
        include_package_data=True,
        author="Ya-Liang Chang",
        author_email="allen314x@yahoo.com.tw",
        description="Bebop Drone control for python",
        platforms="Linux",
        license="BSD",
        keywords="bebop python control",
        url="https://github.com/amjltc295/bybop"
    )



#!/usr/bin/env python3
import sys, os, stat, logging, tempfile, filecmp, shutil
import optparse
import arsdkparser
#===============================================================================
#===============================================================================

#===============================================================================
# Writer class to wrap a file and allow printf-like write method.
#===============================================================================
class Writer(object):
	def __init__(self, filePath):
		# unlink previous file
		try:
			os.unlink(filePath)
		except OSError as e:
			if e.errno == 2:  # No such file or directory
				pass
			else:
				raise e

		self.filePath = filePath
		self.fd = open(filePath, "w")
	def write(self, fmt, *args):
		if args:
			self.fd.write(fmt % (args))
		else:
			self.fd.write(fmt % ())

	def close(self):
		# close file
		self.fd.close()
		# mark file as readonly
		os.chmod(self.filePath, stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH)

#===============================================================================
#===============================================================================
def main():
	(options, args) = parseArgs()
	setupLog(options)
	generator_filepath = args[0]
	extra = ' '.join(args[1:]) if len(args) > 1 else None

	# Setup full path of output directories
	options.outdir = os.path.abspath(options.outdir)

	logging.info("arsdkgen: cmd=%s generator=%s, outdir=%s extra=%s",
		"list" if options.listFiles else "generate",
		generator_filepath, options.outdir, extra)
	if not options.listFiles and not os.path.exists(options.outdir):
		os.makedirs(options.outdir, 0o755)

	# Import package of ge generation
	(generator_path, generator_filename) = os.path.split(generator_filepath)
	sys.path.append(generator_path)
	try:
		generator = __import__(os.path.splitext(generator_filename)[0])
	except ImportError:
		logging.error("Unable to import generator %s", generator_filepath)
		sys.exit(1)

	ctx = arsdkparser.ArParserCtx()
	path, filename = os.path.split(os.path.realpath(__file__))
	path = os.path.join(path, "xml")
	# first load generic.xml
	arsdkparser.parse_xml(ctx, os.path.join(path, "generic.xml"))
	for f in sorted(os.listdir(path)):
		if not f.endswith(".xml") or f == "generic.xml":
			continue
		arsdkparser.parse_xml(ctx, os.path.join(path, f))

	# Finalize features after parsing
	arsdkparser.finalize_ftrs(ctx)

	# Call generator
	if options.listFiles:
		generator.list_files(ctx, options.outdir, extra)
	else:
		# Generate in tmp folder.
		tmp_dir = tempfile.mkdtemp()
		generator.generate_files(ctx, tmp_dir, extra)
		# Copy in outdir all files different in tmp_dir and in outdir.
		for path, subdirs, files in os.walk(tmp_dir):
			for name in files:
				f_tmp = os.path.join(path, name)
				f_cp = f_tmp.replace(tmp_dir, options.outdir)
				if not os.path.isfile(f_cp) or \
						filecmp.cmp(f_tmp, f_cp) == False:
					if not os.path.exists(os.path.dirname(f_cp)):
						os.makedirs(os.path.dirname(f_cp))
					logging.info('Copy: ' + f_tmp + " => " + f_cp)
					shutil.copy(f_tmp, f_cp)
		# Remove tmp_dir
		shutil.rmtree(tmp_dir)

#===============================================================================
# Setup option parser and parse command line.
#===============================================================================
def parseArgs():
	# Setup parser
	usage = "usage: %prog [options] <generator> [extra]"
	parser = optparse.OptionParser(usage = usage)

	# Main options
	parser.add_option("-o", "--output",
		dest = "outdir",
		action = "store",
		default = ".",
		metavar = "DIR",
		help = "Name of output directory [default: current directory]")
	parser.add_option("-f", "--files",
		dest = "listFiles",
		action = "store_true",
		default = False,
		help = "List on stdout full path of files that will be generated and exit")

	# Other options
	parser.add_option("-q",
		dest = "quiet",
		action = "store_true",
		default = False,
		help = "be quiet")
	parser.add_option("-v",
		dest = "verbose",
		action = "count",
		default = 0,
		help = "verbose output (more verbose if specified twice)")

	# Parse arguments and check validity
	(options, args) = parser.parse_args()

	if len(args) < 1:
		parser.error("Missing generator filepath")

	return (options, args)

#===============================================================================
# Setup logging system.
#===============================================================================
def setupLog(options):
	logging.basicConfig(
		level = logging.WARNING,
		format = "[%(levelname)s] %(message)s",
		stream = sys.stderr)
	logging.addLevelName(logging.CRITICAL, "C")
	logging.addLevelName(logging.ERROR, "E")
	logging.addLevelName(logging.WARNING, "W")
	logging.addLevelName(logging.INFO, "I")
	logging.addLevelName(logging.DEBUG, "D")

	# Setup log level
	if options.quiet == True:
		logging.getLogger().setLevel(logging.CRITICAL)
	elif options.verbose >= 2:
		logging.getLogger().setLevel(logging.DEBUG)
	elif options.verbose >= 1:
		logging.getLogger().setLevel(logging.INFO)

#===============================================================================
#===============================================================================
if __name__ == "__main__":
	main()



#!/usr/bin/env python3

import os
import pprint
import xml.dom.minidom

_MIN_PROJECT_ID = 0
_MAX_PROJECT_ID = 255
_MIN_FEATURE_ID = 0
_MAX_FEATURE_ID = 255

_MIN_CLASS_ID = 0
_MAX_CLASS_ID = 255

_MIN_CMD_ID = 0
_MAX_CMD_ID = 65535

_FTR_GEN = 'generic'

#===============================================================================
#===============================================================================
class ArParserError(Exception):
    pass

#===============================================================================
#===============================================================================
class ArCmdListType(object):
    NONE = 0
    LIST = 1
    MAP = 2

    TO_STRING = {NONE: "NONE", LIST: "LIST_ITEM", MAP: "MAP_ITEM"}
    FROM_STRING = {"NONE": NONE, "LIST_ITEM": LIST, "MAP_ITEM": MAP}

#===============================================================================
#===============================================================================
class ArCmdBufferType(object):
    NON_ACK = 0
    ACK = 1
    HIGH_PRIO = 2

    TO_STRING = {NON_ACK: "NON_ACK", ACK: "ACK", HIGH_PRIO: "HIGH_PRIO"}
    FROM_STRING = {"NON_ACK": NON_ACK, "ACK": ACK, "HIGH_PRIO": HIGH_PRIO}

#===============================================================================
#===============================================================================
class ArCmdTimeoutPolicy(object):
    POP = 0
    RETRY = 1

    TO_STRING = {POP: "POP", RETRY: "RETRY"}
    FROM_STRING = {"POP": POP, "RETRY": RETRY}

#===============================================================================
#===============================================================================
class ArCmdContent(object):
    UPDATE = 0
    NOTIFICATION = 1

    TO_STRING = {UPDATE: "UPDATE", NOTIFICATION: "NOTIFICATION"}
    FROM_STRING = {"UPDATE": UPDATE, "NOTIFICATION": NOTIFICATION}

#===============================================================================
#===============================================================================
class ArCmdDeprecation(object):

    TO_STRING = {True: "true", False: "false"}
    FROM_STRING = {"true": True, "false": False}

#===============================================================================
#===============================================================================
class ArArgType(object):
    I8 = 0
    U8 = 1
    I16 = 2
    U16 = 3
    I32 = 4
    U32 = 5
    I64 = 6
    U64 = 7
    FLOAT = 8
    DOUBLE = 9
    STRING = 10
    ENUM = 11
    BITFIELD = 12
    MULTISETTING = 13

    TO_STRING = {I8: "i8", U8: "u8", I16: "i16", U16: "u16",
            I32: "i32", U32: "u32", I64: "i64", U64: "u64",
            FLOAT: "float", DOUBLE: "double", STRING: "string",
            ENUM: "enum", BITFIELD: "bitfield", MULTISETTING: "multisetting"}
    FROM_STRING = {"i8": I8, "u8": U8, "i16": I16, "u16": U16,
            "i32": I32, "u32": U32, "i64": I64, "u64": U64,
            "float": FLOAT, "double": DOUBLE, "string": STRING,
            "enum": ENUM, "bitfield": BITFIELD, "multisetting": MULTISETTING}

#===============================================================================
#===============================================================================
class ArParserCtx(object):
    def __init__(self):
        self.projects = []
        self.projectsById = {}
        self.projectsByName = {}
        self.features = []
        self.featuresById = {}
        self.featuresByName = {}

    def walk_classes(self):
        for projectObj in self.projects:
            for classObj in projectObj.classes:
                yield (projectObj, classObj)

    def walk_cmds(self):
        for projectObj in self.projects:
            for classObj in projectObj.classes:
                for cmdObj in classObj.cmds:
                    yield (projectObj, classObj, cmdObj)

    def walk_msgs(self):
        for featureObj in self.features:
            for msg in featureObj.getMsgs():
                yield (featureObj, msg)

#===============================================================================
#===============================================================================
class ArProject(object):
    def __init__(self, name, projectId, doc):
        self.name = name
        self.projectId = projectId
        self.doc = doc
        self.classes = []
        self.classesById = {}
        self.classesByName = {}

    def __repr__(self):
        return ("{name='%s', projectId=%d, doc='%s', classes=%s}" % (
                self.name,
                self.projectId,
                repr(self.doc),
                pprint.pformat(self.classes)))

#===============================================================================
#===============================================================================
class ArFeature(object):
    def __init__(self, name, featureId, doc):
        self.name = name
        self.featureId = featureId
        self.doc = doc
        self.enums = []
        self.enumsByName = {}
        self.multisets = []
        self.multisetsByName = {}
        self.cmds = []
        self.cmdsById = {} #only for real feature, empty for project
        self.cmdsByName = {} #only for real feature, empty for project
        self.evts = []
        self.evtsById = {}
        self.evtsByName = {}
        self.classes = None #only for project conversion
        self.classesById = {} #only for project conversion
        self.classesByName = {} #only for project conversion

    def getMsgs (self):
        return  self.cmds + self.evts

    def getMsgsById (self):
        #only for feature
        tmp = dict(self.cmdsById)
        tmp.update(self.evtsById)
        return tmp

    def getMsgsByName (self):
        #only for feature
        tmp = dict(self.cmdsByName)
        tmp.update(self.evtsByName)
        return tmp

    def __repr__(self):
        return ("{name='%s', featureId=%d, doc='%s', enums='%s', "
                "multisets='%s', cmds='%s', evts='%s'}" % (
                self.name,
                self.featureId,
                repr(self.doc),
                pprint.pformat(self.enums),
                pprint.pformat(self.multisets),
                pprint.pformat(self.cmds),
                pprint.pformat(self.evts)))

    @staticmethod
    def from_project(prj):
        ftrObj = ArFeature (prj.name, prj.projectId, prj.doc)
        ftrObj.classes = prj.classes
        ftrObj.classesById = prj.classesById
        ftrObj.classesByName = prj.classesByName

        for cl in prj.classes:
            for cmd in cl.cmds:
                msgId = cmd.cmdId
                msgName = cmd.name
                if "event" in cl.name.lower() or "state" in cl.name.lower():
                    msgObj = ArEvt(msgName, msgId, cmd.doc, cmd.listType,
                            cmd.bufferType, cmd.timeoutPolicy, cmd.content,
                            cmd.isDeprecated, ftrObj)
                else:
                    msgObj = ArCmd(msgName, msgId, cmd.doc, cmd.listType,
                            cmd.bufferType, cmd.timeoutPolicy, cmd.content,
                            cmd.isDeprecated, ftrObj)

                if cmd.listType == ArCmdListType.MAP:
                    msgObj.mapKey = cmd.args[0]
                msgObj.cls = cl

                msgObj.args = cmd.args
                msgObj.argsByName = cmd.argsByName
                # Create enums
                for arg in msgObj.args:
                    if len(arg.enums) > 0:
                        enumName = cl.name + '_' +\
                                cmd.name[0].upper()+cmd.name[1:]+'_' +\
                                arg.name[0].upper()+arg.name[1:]
                        enumObj = ArEnum(enumName, arg.doc)
                        enumObj.msg = msgObj
                        for val in arg.enums:
                            eValObj = ArEnumValue(val.name, val.value, val.doc)
                            enumObj.values.append(eValObj)
                            enumObj.valuesByName[val.name] = eValObj
                        ftrObj.enums.append(enumObj)
                        ftrObj.enumsByName[enumName] = enumObj
                        arg.argType = enumObj
                        arg.doc = ''

                if isinstance(msgObj, ArCmd):
                    ftrObj.cmds.append(msgObj)
                else:
                    ftrObj.evts.append(msgObj)
        return ftrObj

#===============================================================================
#===============================================================================
class ArClass(object):
    def __init__(self, name, classId, doc):
        self.name = name
        self.classId = classId
        self.doc = doc
        self.cmds = []
        self.cmdsById = {}
        self.cmdsByName = {}

    def __repr__(self):
        return ("{name='%s', classId=%d, doc='%s', cmds=%s}" % (
                self.name,
                self.classId,
                repr(self.doc),
                pprint.pformat(self.cmds)))

#===============================================================================
#===============================================================================
class ArMsg(object):
    def __init__(self, name, cmdId, doc, listType, bufferType, timeoutPolicy,
            content, isDeprecated, ftr):
        self.name = name
        self.cmdId = cmdId
        self.doc = doc
        self.listType = listType
        self.bufferType = bufferType
        self.timeoutPolicy = timeoutPolicy
        self.content = content
        self.mapKey = None
        self.args = []
        self.argsByName = {}
        self.cls = None #only for project conversion
        self.isDeprecated = isDeprecated
        self.ftr = ftr

    def __repr__(self):
        return ("{name='%s', cmdId=%d, doc='%s', listType='%s', "
                "bufferType='%s', timeoutPolicy='%s', content='%s', "
                "args=%s isDeprecated=%r}" % (
                self.name,
                self.cmdId,
                repr(self.doc),
                ArCmdListType.TO_STRING[self.listType],
                ArCmdBufferType.TO_STRING[self.bufferType],
                ArCmdTimeoutPolicy.TO_STRING[self.timeoutPolicy],
                ArCmdContent.TO_STRING[self.content],
                pprint.pformat(self.args),
                self.isDeprecated))

#===============================================================================
#===============================================================================
class ArCmd(ArMsg):
    def __init__(self, name, cmdId, doc, listType, bufferType, timeoutPolicy,
            content, isDeprecated, ftr):
        ArMsg.__init__(self, name, cmdId, doc, listType, bufferType,
                    timeoutPolicy, content, isDeprecated, ftr)

#===============================================================================
#===============================================================================
class ArEvt(ArMsg):
    def __init__(self, name, cmdId, doc, listType, bufferType, timeoutPolicy,
            content, isDeprecated, ftr):
        ArMsg.__init__(self, name, cmdId, doc, listType, bufferType,
                    timeoutPolicy, content, isDeprecated, ftr)

#===============================================================================
#===============================================================================
class ArComment(object):
    def __init__(self, title, desc, support, triggered, result):
        self.title = title
        self.desc = desc
        self.support = support
        self.triggered = triggered
        self.result = result

    def __repr__(self):
        return ("{title='%s', desc=%s, support='%s', triggered='%s', "
                "result='%s'}" % (
                self.title,
                self.desc,
                self.support,
                self.triggered,
                self.result))

#===============================================================================
#===============================================================================
class ArArg(object):
    def __init__(self, name, argType, doc):
        self.name = name
        self.argType = argType
        self.doc = doc
        self.enums = []
        self.enumsByName = {}

    def __repr__(self):
        if isinstance(self.argType, str):
            argTypeRep = ArArgType.TO_STRING[self.argType]
        else:
            argTypeRep = pprint.pformat(self.argType)

        return ("{name='%s', argType='%s', doc='%s', enums=%s}" % (
                self.name,
                argTypeRep,
                repr(self.doc),
                pprint.pformat(self.enums)))

#===============================================================================
#===============================================================================
class ArMultiSetting(object):
    def __init__(self, name, doc):
        self.name = name
        self.doc = doc
        self.links = []
        self.msgs = []

    def __repr__(self):
        return ("{name='%s', doc='%s', msgs=%s}" % (
                self.name,
                repr(self.doc),
                pprint.pformat(self.msgs)))

#===============================================================================
#===============================================================================
class ArEnumValue(object):
    def __init__(self, name, value, doc):
        self.name = name
        self.doc = doc
        self.value = value

    def __eq__(self, other):
        return self.value == other.value
    def __lt__(self, other):
        return self.value < other.value

    def __repr__(self):
        return ("{name='%s', value=%d, doc='%s'}" % (
                self.name,
                self.value,
                repr(self.doc)))

#===============================================================================
#===============================================================================
class ArEnum(object):
    def __init__(self, name, doc):
        self.name = name
        self.doc = doc
        self.values = []
        self.valuesByName = {}
        self.usedLikeBitfield = False
        self.msg = None #only for project conversion

    def getMaxBitfieldVal(self):
        return 2 ** max(self.values).value

    def __repr__(self):
        return ("{name='%s', doc='%s', values='%s'}" % (
                self.name,
                repr(self.doc),
                pprint.pformat(self.values)))

#===============================================================================
#===============================================================================
class ArBitfield(object):
    TYPE_TO_LENGTH = {ArArgType.U8:2**7, ArArgType.U16:2**15, ArArgType.U32:2**31, ArArgType.U64:2**63}

    def __init__(self, enum, btfType):
        self.enum = enum
        self.btfType = btfType

    def __repr__(self):
        return ("{enum='%s', type='%s'}" % (
                pprint.pformat(self.enum),
                pprint.pformat(self.btfType)))

#===============================================================================
#===============================================================================
def _get_node_content(node):
    try:
        content = node.childNodes[0].nodeValue.strip()
        lines = [l.strip() for l in content.split('\n')]
        return '\n'.join(lines)
    except:
        return ''

#===============================================================================
#===============================================================================
def _parse_project_node(filePath, projectNode, projectObj):
    for classNode in projectNode.getElementsByTagName("class"):
        className = classNode.getAttribute("name")
        classId = int(classNode.getAttribute("id"))
        classDoc = _get_node_content(classNode).strip()

        # Check class id/name
        if classId in projectObj.classesById:
            raise ArParserError("%s: Duplicate class id %d" % (
                    filePath, classId))
        if className in projectObj.classesByName:
            raise ArParserError("%s: Duplicate class name '%s'" % (
                    filePath, className))
        if classId < _MIN_CLASS_ID or classId > _MAX_CLASS_ID:
            raise ArParserError("%s: Invalid class id %d" % (
                    filePath, classId))

        # Create class object
        classObj = ArClass(className, classId, classDoc)
        projectObj.classes.append(classObj)
        projectObj.classesById[classId] = classObj
        projectObj.classesByName[className] = classObj

        # Parse class node
        _parse_class_node(filePath, classNode, classObj)

#===============================================================================
#===============================================================================
def _parse_feature_node(ctx, filePath, featureNode, featureObj):

    for enumsNode in featureNode.getElementsByTagName("enums"):
        for enumNode in enumsNode.getElementsByTagName("enum"):
            enumName = enumNode.getAttribute("name")
            enumDoc = _get_node_content(enumNode).strip()

            # Check enum name
            if enumName in featureObj.enumsByName:
                raise ArParserError("%s: Duplicate enum name '%s'" % (
                        filePath, enumName))

            # Create enum object
            enumObj = ArEnum(enumName, enumDoc)
            featureObj.enums.append(enumObj)
            featureObj.enumsByName[enumName] = enumObj

            # Parse enum node
            _parse_enum_node(filePath, enumNode, enumObj)

    _parse_feature_node_multisets(ctx, filePath, featureNode, featureObj)

    _parse_feature_node_msgs(ctx, filePath, featureNode, featureObj)

#===============================================================================
#===============================================================================
def _parse_feature_node_multisets(ctx, filePath, featureNode, featureObj):
    for multisetsNode in featureNode.getElementsByTagName("multisettings"):
        for multisetNode in multisetsNode.getElementsByTagName("multisetting"):
            multisetName = multisetNode.getAttribute("name")
            multisetDoc = _get_node_content(multisetNode).strip()

            # Check multiset name
            if multisetName in featureObj.multisetsByName:
                raise ArParserError("%s: Duplicate multiset name '%s'" % (
                        filePath, multisetName))

            # Create multiset object
            multisetObj = ArMultiSetting(multisetName, multisetDoc)
            featureObj.multisets.append(multisetObj)
            featureObj.multisetsByName[multisetName] = multisetObj

            # Parse multiset node
            _parse_multiset_node(filePath, multisetNode, multisetObj)

#===============================================================================
#===============================================================================
def _parse_multiset_node(filePath, multisetNode, multisetObj):
    for memberNode in multisetNode.getElementsByTagName("member"):
        multisetObj.links.append(memberNode.getAttribute("link"))

#===============================================================================
#===============================================================================
def _parse_feature_node_msgs(ctx, filePath, featureNode, featureObj):

    for msgsNode in featureNode.getElementsByTagName("msgs"):
        for msgNode in msgsNode.getElementsByTagName("cmd") + \
                msgsNode.getElementsByTagName("evt"):
            msgName = msgNode.getAttribute("name")
            msgId = int(msgNode.getAttribute("id"))
            msgDoc = _get_node_content(msgNode).strip()

            if msgId < _MIN_CMD_ID or msgId > _MAX_CMD_ID:
                raise ArParserError("%s: Invalid msg id %d" % (
                        filePath, msgId))

            # Check msg name
            if msgName in featureObj.getMsgsByName():
                raise ArParserError("%s: Duplicate message name '%s'" % (
                        filePath, msgName))

            # Check msg id
            if msgId in featureObj.getMsgsById():
                raise ArParserError("%s: Duplicate message id '%s'" % (
                        filePath, msgName))

            # Get type
            msgType = ArCmdListType.NONE
            mapKey = None
            if msgNode.hasAttribute("type"):
                attr, _, mapKey = msgNode.getAttribute("type").partition(':')

                if attr not in ArCmdListType.FROM_STRING:
                    raise ArParserError("%s: Invalid list type '%s'" % (
                            filePath, attr))
                msgType = ArCmdListType.FROM_STRING[attr]

            # Get buffer type
            msgBufferType = ArCmdBufferType.ACK
            if msgNode.hasAttribute("buffer"):
                attr = msgNode.getAttribute("buffer")
                if attr not in ArCmdBufferType.FROM_STRING:
                    raise ArParserError("%s: Invalid buffer type '%s'" % (
                            filePath, attr))
                msgBufferType = ArCmdBufferType.FROM_STRING[attr]

            # Get timeout policy
            msgTimeoutPolicy = ArCmdTimeoutPolicy.POP
            if msgNode.hasAttribute("timeout"):
                attr = msgNode.getAttribute("timeout")
                if attr not in ArCmdTimeoutPolicy.FROM_STRING:
                    raise ArParserError("%s: Invalid timout policy '%s'" % (
                            filePath, attr))
                msgTimeoutPolicy = ArCmdTimeoutPolicy.FROM_STRING[attr]

            # Get Content
            msgContent = ArCmdContent.UPDATE
            if msgNode.hasAttribute("content"):
                attr = msgNode.getAttribute("content")
                if attr not in ArCmdContent.FROM_STRING:
                    raise ArParserError("%s: Invalid notification '%s'" % (
                            filePath, attr))
                msgContent = ArCmdContent.FROM_STRING[attr]

            # Get if the message is deprecated
            mgsIsDeprecated = False
            if msgNode.hasAttribute("deprecated"):
                attr = msgNode.getAttribute("deprecated")
                if attr not in ArCmdDeprecation.FROM_STRING:
                    mgsIsDeprecated = ArCmdDeprecation.FROM_STRING[attr]

            # Create msg object
            if msgNode in msgsNode.getElementsByTagName("cmd"):
                #is command
                msgObj = ArCmd (msgName, msgId, msgDoc,
                        msgType, msgBufferType, msgTimeoutPolicy, msgContent,
                        mgsIsDeprecated, featureObj)

            else:
                #is event
                msgObj = ArEvt(msgName, msgId, msgDoc,
                        msgType, msgBufferType, msgTimeoutPolicy, msgContent,
                        mgsIsDeprecated, featureObj)

            # Parse msg node
            _parse_msg_node(ctx, filePath, featureObj, msgNode, msgObj)

            # Find map key
            if mapKey :
                if mapKey not in msgObj.argsByName:
                    raise ArParserError("%s: Invalid Map Key '%s'" % (
                            filePath, mapKey))
                msgObj.mapKey = msgObj.argsByName[mapKey]

            if isinstance(msgObj, ArCmd):
                featureObj.cmds.append(msgObj)
                featureObj.cmdsById[msgId] = msgObj
                featureObj.cmdsByName[msgName] = msgObj
            else:
                featureObj.evts.append(msgObj)
                featureObj.evtsById[msgId] = msgObj
                featureObj.evtsByName[msgName] = msgObj

#===============================================================================
#===============================================================================
def _parse_class_node(filePath, classNode, classObj):
    for cmdNode in classNode.getElementsByTagName("cmd"):
        cmdName = cmdNode.getAttribute("name")
        cmdId = int(cmdNode.getAttribute("id"))
        cmdDoc = _get_cmt_node(cmdNode)

        if cmdId < _MIN_CMD_ID or cmdId > _MAX_CMD_ID:
            raise ArParserError("%s: Invalid cmd id %d" % (
                    filePath, cmdId))

        # Get list type
        cmdListType = ArCmdListType.NONE
        if cmdNode.hasAttribute("type"):
            attr = cmdNode.getAttribute("type")
            if attr not in ArCmdListType.FROM_STRING:
                raise ArParserError("%s: Invalid list type '%s'" % (
                        filePath, attr))
            cmdListType = ArCmdListType.FROM_STRING[attr]

        # Get buffer type
        cmdBufferType = ArCmdBufferType.ACK
        if cmdNode.hasAttribute("buffer"):
            attr = cmdNode.getAttribute("buffer")
            if attr not in ArCmdBufferType.FROM_STRING:
                raise ArParserError("%s: Invalid buffer type '%s'" % (
                        filePath, attr))
            cmdBufferType = ArCmdBufferType.FROM_STRING[attr]

        # Get timeout policy
        cmdTimeoutPolicy = ArCmdTimeoutPolicy.POP
        if cmdNode.hasAttribute("timeout"):
            attr = cmdNode.getAttribute("timeout")
            if attr not in ArCmdTimeoutPolicy.FROM_STRING:
                raise ArParserError("%s: Invalid timout policy '%s'" % (
                        filePath, attr))
            cmdTimeoutPolicy = ArCmdTimeoutPolicy.FROM_STRING[attr]

        # Check cmd name
        if cmdName in classObj.cmdsByName:
            raise ArParserError("%s: Duplicate cmd name '%s'" % (
                    filePath, cmdName))

        # Get cmd Content
        cmdContent = ArCmdContent.UPDATE
        if cmdNode.hasAttribute("content"):
            attr = cmdNode.getAttribute("content")
            if attr not in ArCmdContent.FROM_STRING:
                raise ArParserError("%s: Invalid notification '%s'" % (
                        filePath, attr))
            cmdContent = ArCmdContent.FROM_STRING[attr]

        # Get if the message is deprecated
        mgsIsDeprecated = False
        if cmdNode.hasAttribute("deprecated"):
            attr = cmdNode.getAttribute("deprecated")
            if attr == "true":
                mgsIsDeprecated = True

        # Create cmd object
        cmdObj = ArCmd(cmdName, cmdId, cmdDoc, cmdListType, cmdBufferType,
                    cmdTimeoutPolicy, cmdContent, mgsIsDeprecated, None)
        cmdObj.cls = classObj
        classObj.cmds.append(cmdObj)
        classObj.cmdsById[cmdId] = cmdObj
        classObj.cmdsByName[cmdName] = cmdObj

        # Parse cmd node
        _parse_prj_cmd_node(filePath, cmdNode, cmdObj)

#===============================================================================
#===============================================================================
def _parse_prj_cmd_node(filePath, cmdNode, cmdObj):
    for argNode in cmdNode.getElementsByTagName("arg"):
        argName = argNode.getAttribute("name")
        argDoc = _get_node_content(argNode).strip()

        # Arg type
        attr = argNode.getAttribute("type")
        if attr not in ArArgType.FROM_STRING:
            raise ArParserError("%s: Invalid arg type '%s'" % (
                    filePath, attr))
        argType = ArArgType.FROM_STRING[attr]

        # Check arg name
        if argName in cmdObj.argsByName:
            raise ArParserError("%s: Duplicate arg name '%s'" % (
                    filePath, argName))

        # Create arg object
        argObj = ArArg(argName, argType, argDoc)
        cmdObj.args.append(argObj)
        cmdObj.argsByName[argName] = argObj

        # Parse arg node
        _parse_arg_node(filePath, argNode, argObj)

def _fmt_cmt_node(raw_cmt):
    one_line = ' '.join(raw_cmt.split())
    lines = [l.strip() for l in one_line.split(r'\n')]
    res = '\n'.join(lines)
    return res

def _get_cmt_node(msgNode):
    if msgNode.getElementsByTagName("comment"):
        commentNode = msgNode.getElementsByTagName("comment")[0]
        cmtTitle = commentNode.getAttribute("title")
        cmtSupport = commentNode.getAttribute("support")

        cmtDesc = _fmt_cmt_node(commentNode.getAttribute("desc"))

        if commentNode.hasAttribute("triggered"):
            cmtTriggered = _fmt_cmt_node(commentNode.getAttribute("triggered"))
        else:
            cmtTriggered = None

        if commentNode.hasAttribute("result"):
            cmtResult = _fmt_cmt_node(commentNode.getAttribute("result"))
        else:
            cmtResult = None

        # Create comment object
        return ArComment(cmtTitle, cmtDesc, cmtSupport,
                cmtTriggered, cmtResult)
    else:
        oldComment = _get_node_content(msgNode)
        return ArComment(oldComment.splitlines()[0], oldComment, None,
                None, None)

#===============================================================================
#===============================================================================
def _parse_msg_node(ctx, filePath, ftr, msgNode, msgObj):
    if msgNode.getElementsByTagName("comment"):
        commentNode = msgNode.getElementsByTagName("comment")[0]
        cmtTitle = commentNode.getAttribute("title")
        cmtSupport = commentNode.getAttribute("support")

        cmtDesc = _fmt_cmt_node(commentNode.getAttribute("desc"))

        if commentNode.hasAttribute("triggered"):
            cmtTriggered = _fmt_cmt_node(commentNode.getAttribute("triggered"))
        else:
            cmtTriggered = None

        if commentNode.hasAttribute("result"):
            cmtResult = _fmt_cmt_node(commentNode.getAttribute("result"))
        else:
            cmtResult = None

        # Create comment object
        msgObj.doc = ArComment(cmtTitle, cmtDesc, cmtSupport,
                cmtTriggered, cmtResult)
    else:
        oldComment = _get_node_content(msgNode)
        msgObj.doc = ArComment(oldComment.splitlines()[0], oldComment, None,
                None, None)

    _parse_msg_node_args(ctx, filePath, ftr, msgNode, msgObj)

#===============================================================================
#===============================================================================
def _parse_msg_node_args(ctx, filePath, ftr, msgNode, msgObj):
    for argNode in msgNode.getElementsByTagName("arg"):
        argName = argNode.getAttribute("name")
        argDoc = _get_node_content(argNode).strip()

        # Get type attrs
        attr1, _, flw = argNode.getAttribute("type").partition(':')
        attr2, _, attr3 = flw.partition(':')
        # Check arg type
        if attr1 not in ArArgType.FROM_STRING:
            raise ArParserError("%s: Invalid arg type '%s'" % (
                    filePath, attr1))

        if ArArgType.FROM_STRING[attr1] == ArArgType.ENUM:
            # Find Enum
            if attr2 not in ftr.enumsByName and \
                    (_FTR_GEN not in ctx.featuresByName or \
                    attr2 not in ctx.featuresByName[_FTR_GEN].enumsByName):
                raise ArParserError("%s: Invalid enum arg type '%s'" % (
                    filePath, attr2))

            if attr2 in ftr.enumsByName:
                argType = ftr.enumsByName[attr2]
            else:
                argType = ctx.featuresByName[_FTR_GEN].enumsByName[attr2]

        elif ArArgType.FROM_STRING[attr1] == ArArgType.BITFIELD:
            # Find Enum
            if attr3 not in ftr.enumsByName and \
                    (_FTR_GEN not in ctx.featuresByName or \
                    attr3 not in ctx.featuresByName[_FTR_GEN].enumsByName):
                raise ArParserError("%s: Invalid bitfield enum arg type '%s'"
                        % (filePath, attr3))

            # Check bitfield length
            if attr2 not in ArArgType.FROM_STRING and \
                    ArArgType.FROM_STRING[attr2] in ArBitfield.TYPE_TO_LENGTH:
                raise ArParserError("%s: Invalid bitfield enum arg length '%s'"
                        % (filePath, attr2))

            if attr3 in ftr.enumsByName:
                btfEnum = ftr.enumsByName[attr3]
            else:
                btfEnum = ctx.featuresByName[_FTR_GEN].enumsByName[attr3]
            btfType = ArArgType.FROM_STRING[attr2]

            # Check Compatibility between Enum max value and bitfield length
            if ArBitfield.TYPE_TO_LENGTH[btfType] < btfEnum.getMaxBitfieldVal():
                raise ArParserError("%s: Too Small bitfield length '%s.%s'"
                        % (filePath, msgObj.name, argName))

            argType = ArBitfield(btfEnum, btfType)
            btfEnum.usedLikeBitfield = True
        elif ArArgType.FROM_STRING[attr1] == ArArgType.MULTISETTING:
            # Find multi setting
            if attr2 not in ftr.multisetsByName and \
                    (_FTR_GEN not in ctx.featuresByName or \
                    attr2 not in ctx.featuresByName[_FTR_GEN].multisetsByName):
                raise ArParserError("%s: Invalid multisetting arg type '%s'"
                        % (filePath, attr2))

            if attr2 in ftr.multisetsByName:
                argType = ftr.multisetsByName[attr2]
            else:
                argType = ctx.featuresByName[_FTR_GEN].multisetsByName[attr2]
        else:
            argType = ArArgType.FROM_STRING[attr1]

        # Check arg name
        if argName in msgObj.argsByName:
            raise ArParserError("%s: Duplicate arg name '%s'" % (
                    filePath, argName))

        # Create arg object
        argObj = ArArg(argName, argType, argDoc)
        msgObj.args.append(argObj)
        msgObj.argsByName[argName] = argObj

        # Parse arg node
        _parse_arg_node(filePath, argNode, argObj)

#===============================================================================
#===============================================================================
def _parse_arg_node(filePath, argNode, argObj):
    nextValue = 0
    for enumNode in argNode.getElementsByTagName("enum"):
        enumName = enumNode.getAttribute("name")
        enumDoc = _get_node_content(enumNode).strip()

        enumValue = nextValue
        nextValue += 1

        # Check enum name
        if enumName in argObj.enumsByName:
            raise ArParserError("%s: Duplicate enum name '%s'" % (
                    filePath, enumName))

        # Create enum object
        enumObj = ArEnumValue(enumName, enumValue, enumDoc)
        argObj.enums.append(enumObj)
        argObj.enumsByName[enumName] = enumObj

#===============================================================================
#===============================================================================
def _parse_enum_node(filePath, enumNode, enumObj):
    nextValue = 0
    for eValNode in enumNode.getElementsByTagName("value"):
        eValName = eValNode.getAttribute("name")
        eValDoc = _get_node_content(eValNode).strip()

        if eValNode.hasAttribute("val"):
            eValVal = int(eValNode.getAttribute("val"))
        else:
            eValVal = nextValue
            nextValue += 1
        nextValue = eValVal + 1

        # Check enum value name
        if eValName in enumObj.valuesByName:
            raise ArParserError("%s: Duplicate enum value name '%s'" % (
                    filePath, eValName))

        # Create enum value object
        eValObj = ArEnumValue(eValName, eValVal, eValDoc)
        enumObj.values.append(eValObj)
        enumObj.valuesByName[eValName] = eValObj

#===============================================================================
#===============================================================================
def parse_prj_xml(ctx, filePath):
    # Parse project xml file
    try:
        xmlDom = xml.dom.minidom.parse(filePath)
    except Exception as ex:
        raise ArParserError("Error while loading '%s': %s" % (
                filePath, str(ex)))

    # Get project node
    projectNode = xmlDom.documentElement
    if projectNode.tagName != "project":
        raise ArParserError("%s: Bad root element: '%s'" % (
                filePath, projectNode.tagName))
    projectName = projectNode.getAttribute("name")
    projectId = int(projectNode.getAttribute("id"))
    projectDoc = _get_node_content(projectNode).strip()

    # Check project id/name
    if projectId in ctx.projectsById:
        raise ArParserError("%s: Duplicate project id %d" % (
                filePath, projectId))
    if projectId < _MIN_PROJECT_ID or projectId > _MAX_PROJECT_ID:
        raise ArParserError("%s: Invalid project id %d" % (
                filePath, projectId))
    if projectName in ctx.projectsByName:
        raise ArParserError("%s: Duplicate project name '%s'" % (
                filePath, projectName))

    # Create project object
    projectObj = ArProject(projectName, projectId, projectDoc)
    ctx.projects.append(projectObj)
    ctx.projectsById[projectId] = projectObj
    ctx.projectsByName[projectName] = projectObj

    # Parse project node
    _parse_project_node(filePath, projectNode, projectObj)

    # Convert project to feature object
    featureObj = ArFeature.from_project(projectObj)
    ctx.features.append(featureObj)
    ctx.featuresById[featureObj.featureId] = featureObj
    ctx.featuresByName[featureObj.name] = featureObj

#===============================================================================
#===============================================================================
def parse_ftr_xml(ctx, filePath):
    # Parse feature xml file
    try:
        xmlDom = xml.dom.minidom.parse(filePath)
    except Exception as ex:
        raise ArParserError("Error while loading '%s': %s" % (
                filePath, str(ex)))

    # Get feature node
    featureNode = xmlDom.documentElement
    if featureNode.tagName != "feature":
        raise ArParserError("%s: Bad root element: '%s'" % (
                filePath, featureNode.tagName))
    featureName = featureNode.getAttribute("name")
    featureId = int(featureNode.getAttribute("id"))
    featureDoc = _get_node_content(featureNode).strip()

    # Check feature id/name
    if featureId in ctx.featuresById:
        raise ArParserError("%s: Duplicate feature id %d" % (
                filePath, featureId))
    if featureId < _MIN_FEATURE_ID or featureId > _MAX_FEATURE_ID:
        raise ArParserError("%s: Invalid feature id %d" % (
                filePath, featureId))
    if featureName in ctx.featuresByName:
        raise ArParserError("%s: Duplicate feature name '%s'" % (
                filePath, featureName))

    # Create feature object
    featureObj = ArFeature(featureName, featureId, featureDoc)
    ctx.features.append(featureObj)
    ctx.featuresById[featureId] = featureObj
    ctx.featuresByName[featureName] = featureObj

    # Parse feature node
    _parse_feature_node(ctx, filePath, featureNode, featureObj)

#===============================================================================
#===============================================================================
def _link_to_msg(ctx, link):
    parts = link.split(".")

    if len(parts) < 2:
        return None

    if not parts[0] in ctx.featuresByName:
        return None
    ftr = ctx.featuresByName[parts[0]]

    if len(parts) == 2:
        return ftr.getMsgsByName[parts[1]]

    # Project part
    clsName = parts[1]
    cmdName = parts[2]

    for cmd in ftr.cmds + ftr.evts:
        if cmd.name == cmdName and cmd.cls and cmd.cls.name == clsName:
            return cmd

    return None

#===============================================================================
#===============================================================================
def finalize_ftrs(ctx):
    # Finalize features
    for ftr in ctx.features:
        # Finalize multi settings
        for multiset in ftr.multisets:
            for link in multiset.links:
                msg = _link_to_msg(ctx, link)
                if not msg:
                    raise ArParserError("%s: Bad multisetting link '%s'" % (
                            filePath, link))
                multiset.msgs.append(msg)

#===============================================================================
#===============================================================================
def parse_xml(ctx, filePath):
    # Parse xml file
    try:
        xmlDom = xml.dom.minidom.parse(filePath)
    except Exception as ex:
        raise ArParserError("Error while loading '%s': %s" % (
                filePath, str(ex)))

    # Get feature node
    node = xmlDom.documentElement
    if node.tagName == "feature":
        parse_ftr_xml(ctx, filePath)
    elif node.tagName == "project":
        parse_prj_xml(ctx, filePath)

#===============================================================================
#===============================================================================
def main():
    ctx = ArParserCtx()
    path, filename = os.path.split(os.path.realpath(__file__))
    path = os.path.join(path, "xml")

    # first load generic.xml
    parse_xml(ctx, os.path.join(path, "generic.xml"))
    for f in sorted(os.listdir(path)):
        if not f.endswith(".xml") or f == "generic.xml":
            continue
        parse_xml(ctx, os.path.join(path, f))

    # Finalize MultiSettings
    finalize_ftrs(ctx)

    #for prj in ctx.projects:
    #    print prj
    #    print '\n'
    #for f in ctx.features:
    #    print f
    #    print '\n'

#===============================================================================
#===============================================================================
if __name__ == "__main__":
    main()



import os
import sys
import struct

MY_PATH, _ = os.path.split(os.path.realpath(__file__))
ARSDK_PATH=os.path.join(MY_PATH,'..', 'arsdk-xml')
ARCOMMANDS_PATH=os.path.join(ARSDK_PATH, 'xml')

sys.path.append(ARSDK_PATH)

import arsdkparser

_ctx = arsdkparser.ArParserCtx()
arsdkparser.parse_xml(_ctx, os.path.join(ARCOMMANDS_PATH, 'generic.xml'))
for f in sorted(os.listdir(ARCOMMANDS_PATH)):
    if not f.endswith('.xml') or f == 'generic.xml':
        continue
    arsdkparser.parse_xml(_ctx, os.path.join(ARCOMMANDS_PATH, f))
arsdkparser.finalize_ftrs(_ctx)


class CommandError(Exception):
    def __init__(self, msg):
        self.value = msg
    def __str__(self):
        return repr(self.value)

_struct_fmt_for_type = {
    'u8'     : 'B',
    'i8'     : 'b',
    'u16'    : 'H',
    'i16'    : 'h',
    'u32'    : 'I',
    'i32'    : 'i',
    'u64'    : 'Q',
    'i64'    : 'q',
    'float'  : 'f',
    'double' : 'd',
    'string' : 'z',
    'enum'   : 'i',
}

def _format_string_for_cmd(cmd):
    ret = '<'
    for arg in cmd.args:
        if isinstance(arg.argType, arsdkparser.ArMultiSetting):
            raise Exception('Multisettings not supported !')
        elif isinstance(arg.argType, arsdkparser.ArBitfield):
            arg_str_type = arsdkparser.ArArgType.TO_STRING[arg.argType.btfType]
        elif isinstance(arg.argType, arsdkparser.ArEnum):
            arg_str_type = 'i32'
        else:
            arg_str_type = arsdkparser.ArArgType.TO_STRING[arg.argType]
        ret += _struct_fmt_for_type[arg_str_type]
    return ret, bool(cmd.args)

def _struct_pack(fmt, *args):
    """
    like struct.pack(fmt, *args)
    except that a 'z' format is supported to include null terminated strings
    """
    nbarg = 0
    real_fmt = ''
    for c in fmt:
        if c == 'z':
            real_fmt += '%ds' % (len(args[nbarg])+1)
            nbarg += 1
        else:
            real_fmt += c
            if c in 'cbB?hHiIlLqQfdspP':
                nbarg += 1
    return struct.pack(real_fmt, *args)


def _struct_unpack(fmt, string):
    """
    like struct.unpack(fmt, string)
    except that a 'z' format is supported to read a null terminated string
    """
    real_fmt=''
    null_idx=[]
    nbarg = 0
    for i in range(len(fmt)):
        c = fmt[i]
        if c == 'z':
            start = struct.calcsize(real_fmt)
            strlen = string[start:].find('\0')
            if strlen < 0:
                raise CommandError('No null char in string')
            real_fmt += '%dsB' % strlen
            nbarg += 1
            null_idx.append(nbarg)
            nbarg += 1
        else:
            real_fmt += c
            if c in 'cbB?hHiIlLqQfdspP':
                nbarg += 1

    content = struct.unpack(real_fmt, string)
    ret = tuple([content[i] for i in range(len(content)) if i not in null_idx])
    return ret

def pack_command(s_proj, s_cls, s_cmd, *args):
    """
    Pack a command into a string.

    Arguments:
    - s_proj : Name of the project
    - s_cls  : Name of the class within the project (ignored for features)
    - s_cmd  : Name of the command within the class
    - *args  : Arguments of the command.

    If the project, the class or the command can not be found in the command table,
    a CommandError will be raised.

    If the number and type of arguments in *arg do not match the expected ones, a
    CommandError will be raised.

    Return the command string, the command recommanded buffer and the command
    recommanded timeout policy.
    """
    proj = None
    feat = None
    projid = 0
    cls = None
    clsid = 0
    cmd = None
    # Let an exception be raised if we do not know the command or if the format is bad
    # Find the project
    if s_proj in _ctx.projectsByName:
        proj = _ctx.projectsByName[s_proj]
    elif s_proj in _ctx.featuresByName:
        feat = _ctx.featuresByName[s_proj]
    if proj is None and feat is None:
        raise CommandError('Unknown project ' + s_proj)

    if proj: # Project
        projid = proj.projectId
        # Find the class
        if s_cls in proj.classesByName:
            cls = proj.classesByName[s_cls]
        if cls is None:
            raise CommandError('Unknown class ' + s_cls + ' in project ' + s_proj)
        clsid = cls.classId

        # Find the command
        if s_cmd in cls.cmdsByName:
            cmd = cls.cmdsByName[s_cmd]
        if cmd is None:
            raise CommandError('Unknown command ' + s_cmd + ' in class ' + s_cls + ' of project ' + s_proj)
    elif feat: # Feature
        projid = feat.featureId
        # Find the command
        if s_cmd in feat.cmdsByName:
            cmd = feat.cmdsByName[s_cmd]
        if cmd is None:
            raise CommandError('Unknown command ' + s_cmd + ' in feature ' + s_proj)

    ret = struct.pack('<BBH', projid, clsid, cmd.cmdId)
    argsfmt, needed = _format_string_for_cmd(cmd)
    if needed:
        try:
            ret += _struct_pack(argsfmt, *args)
        except IndexError:
            raise CommandError('Missing arguments')
        except TypeError:
            raise CommandError('Bad type for arguments')
        except struct.error:
            raise CommandError('Bad type for arguments')


    return ret, cmd.bufferType, cmd.timeoutPolicy

def unpack_command(buf):
    """
    Unpack a command string into a dictionnary of arguments

    Arguments:
    - buf : The packed command

    Return a dictionnary describing the command, and a boolean indicating whether the
    command is known. If the boolean is False, then the dictionnary is {}

    Return dictionnary format:
    {
      'name' : full name of the command (project.class.command)
      'project' : project of the command
      'class' : class of the command
      'cmd' : command name
      'listtype' : list type (none/list/map) of the command
      'args' : arguments in the commands, in the form { 'name':value, ... }
      'arg0' : value of the first argument ('' if no arguments)
               this is useful for map commands, as this will be the key.
    }

    A CommandError is raised if the command is in a bad format.
    """
    # Read the project/cls/cmd from the buffer
    try:
        (i_proj, i_cls, i_cmd) = struct.unpack('<BBH', buf[:4])
    except struct.error:
        raise CommandError('Bad input buffer (not an ARCommand)')
    proj = None
    feat = None
    cls = None
    cmd = None
    # Let an exception be raised if we do not know the command or if the format is bad

    # Find the project
    if i_proj in _ctx.projectsById:
        proj = _ctx.projectsById[i_proj]
    # Or the feature
    if i_proj in _ctx.featuresById:
        feat = _ctx.featuresById[i_proj]

    # If project, Find the class
    if proj:
        if i_cls in proj.classesById:
            cls = proj.classesById[i_cls]
        else:
            return {}, False

        if i_cmd in cls.cmdsById:
            cmd = cls.cmdsById[i_cmd]
        else:
            return {}, False
    # If feature, find directly the command
    elif feat:
        if i_cmd in feat.cmdsById:
            cmd = feat.cmdsById[i_cmd]
        elif i_cmd in feat.evtsById:
            cmd = feat.evtsById[i_cmd]
        else:
            return {}, False
    else:
        return {}, False

    args = ()
    argsfmt, needed = _format_string_for_cmd(cmd)
    if needed:
        try:
            args = _struct_unpack(argsfmt, buf[4:])
        except struct.error:
            raise CommandError('Bad input buffers (arguments do not match the command)')

    ret = {
        'name'     : '%s.%s.%s' % (proj.name if proj else feat.name, cls.name if cls else '', cmd.name),
        'proj'     : proj.name if proj else feat.name,
        'class'    : cls.name if cls else '',
        'cmd'      : cmd.name,
        'listtype' : cmd.listType,
        'listtype_str' : arsdkparser.ArCmdListType.TO_STRING[cmd.listType],
        'args'     : {},
        'arg0'     : '',
    }

    for i in range(len(args)):
        if i == 0:
            ret['arg0'] = args[0]
        ret['args'][cmd.args[i].name] = args[i]

    return ret, True
